```

    Inverse problem script to recover geodesic equations from waveform data

```
cd(@__DIR__)
using Pkg; Pkg.activate("."); Pkg.instantiate()
using OrdinaryDiffEq
using DiffEqFlux
using Optim
using LineSearches
using DataFrames
using CSV
using Plots
gr()

include("utils.jl")
include("models.jl")

## Define the experimental parameters
mass_ratio = 0.0
u0 = Float64[pi, 0.0]
datasize = 250
tspan = (0.0f0, 6.0f4)
tsteps = range(tspan[1], tspan[2], length = datasize)
dt_data = tsteps[2] - tsteps[1]
dt = 100.0
model_params = [100.0, 1.0, 0.5] # p, M, e

# Generate waveform data
prob = ODEProblem(RelativisticOrbitModel, u0, tspan, model_params)
soln = Array(solve(prob, RK4(), saveat = tsteps, dt = dt, adaptive=false))
waveform = compute_waveform(dt_data, soln, mass_ratio, model_params)[1]

plt = plot(tsteps, waveform,
           markershape=:circle, markeralpha = 0.25,
           linewidth = 2, alpha = 0.5,
           label="waveform data")

## Define neural network model
NN = FastChain((x, p) -> [cos(x[1])],
                FastDense(1, 32, cos),
                FastDense(32, 32, cos),
                FastDense(32, 2))
NN_params = initial_params(NN) .* 0

function ODE_model(u, NN_params, t)
    du = AbstractNNOrbitModel(u, model_params, t, NN=NN, NN_params=NN_params)
    return du
end

prob_nn = ODEProblem(ODE_model, u0, tspan, NN_params)
soln_nn = Array(solve(prob_nn, RK4(), u0 = u0, p = NN_params, saveat = tsteps, dt = dt, adaptive=false))
waveform_nn = compute_waveform(dt_data, soln_nn, mass_ratio, model_params)[1]

plot!(plt, tsteps, waveform_nn,
           markershape=:circle, markeralpha = 0.25,
           linewidth = 2, alpha = 0.5,
           label="waveform NN")
display(plt)

## Define objective function

function loss(NN_params)
    pred = Array(solve(prob_nn, RK4(), u0 = u0, p = NN_params, saveat = tsteps, dt = dt, adaptive=false))
    pred_waveform = compute_waveform(dt_data, pred, mass_ratio, model_params)[1]

    loss = ( sum(abs2, waveform .- pred_waveform) )
    return loss, pred_waveform
end

# Test loss function
loss(NN_params)

const losses = []

callback(θ,l,pred_waveform; doplot = true) = begin
    push!(losses, l)
    display(l)
    # plot current prediction against data
    plt = plot(tsteps, waveform,
        markershape=:circle, markeralpha = 0.25,
        linewidth = 2, alpha = 0.5,
        label="wform data (h22)", legend=:topleft)
    plot!(plt, tsteps, pred_waveform,
        markershape=:circle, markeralpha = 0.25,
        linewidth = 2, alpha = 0.5,
        label = "wform NN")
    if doplot
        display(plot(plt))
    end
    # Tell sciml_train to not halt the optimization. If return true, then
    # optimization stops.
    return false
end

## Train with BFGS (gives best results because the Newtonian model seems to give a very good initial guess)
NN_params = NN_params + Float64(1e-3)*randn(eltype(NN_params), size(NN_params))
res = DiffEqFlux.sciml_train(loss, NN_params, BFGS(initial_stepnorm=0.01, linesearch = LineSearches.BackTracking()), cb=callback, maxiters = 20)

## Plot learned solutions

reference_solution = solve(remake(prob, p = model_params, saveat = tsteps, tspan=tspan),
                            RK4(), dt = dt, adaptive=false)
optimized_solution = solve(remake(prob_nn, p = res.minimizer, saveat = tsteps, tspan=tspan),
                            RK4(), dt = dt, adaptive=false)
Newtonian_prob = ODEProblem(NewtonianOrbitModel, u0, tspan, model_params)
Newtonian_solution = solve(remake(Newtonian_prob, p = model_params, saveat = tsteps, tspan=tspan),
                            RK4(), dt = dt, adaptive=false)
true_orbit = soln2orbit(reference_solution, model_params)
pred_orbit = soln2orbit(optimized_solution, model_params)
Newt_orbit = soln2orbit(Newtonian_solution, model_params)

df_learned_trajectories = DataFrame(time = tsteps,
                       true_orbit_x = true_orbit[1,:],
                       true_orbit_y = true_orbit[2,:],
                       pred_orbit_x = pred_orbit[1,:],
                       pred_orbit_y = pred_orbit[2,:],
                       Newt_orbit_x = Newt_orbit[1,:],
                       Newt_orbit_y = Newt_orbit[2,:])

true_waveform = compute_waveform(dt_data, reference_solution, mass_ratio, model_params)[1]
pred_waveform = compute_waveform(dt_data, optimized_solution, mass_ratio, model_params)[1]
Newt_waveform = compute_waveform(dt_data, Newtonian_solution, mass_ratio, model_params)[1]

df_learned_waveforms = DataFrame(time = tsteps,
                                true_waveform = true_waveform,
                                pred_waveform = pred_waveform,
                                Newt_waveform = Newt_waveform)

## Plot extrapolated solutions
factor=5

extended_tspan = (tspan[1], factor*tspan[2])
extended_tsteps = range(tspan[1], factor*tspan[2], length = factor*datasize)
reference_solution = solve(remake(prob, p = model_params, saveat = extended_tsteps, tspan=extended_tspan),
                            RK4(), dt = dt, adaptive=false)
optimized_solution = solve(remake(prob_nn, p = res.minimizer, saveat = extended_tsteps, tspan=extended_tspan),
                            RK4(), dt = dt, adaptive=false)
Newtonian_prob = ODEProblem(NewtonianOrbitModel, u0, tspan, model_params)
Newtonian_solution = solve(remake(Newtonian_prob, p = model_params, saveat = extended_tsteps, tspan=extended_tspan),
                            RK4(), dt = dt, adaptive=false)
true_orbit = soln2orbit(reference_solution, model_params)
pred_orbit = soln2orbit(optimized_solution, model_params)
Newt_orbit = soln2orbit(Newtonian_solution, model_params)
plt = plot(true_orbit[1,:], true_orbit[2,:], linewidth = 2, label = "truth")
plot!(plt, pred_orbit[1,:], pred_orbit[2,:], linestyle = :dash, linewidth = 2, label = "prediction")
plot!(plt, Newt_orbit[1,:], Newt_orbit[2,:], linewidth = 2, label = "Newtonian")
display(plot(plt))

true_waveform = compute_waveform(dt_data, reference_solution, mass_ratio, model_params)[1]
pred_waveform = compute_waveform(dt_data, optimized_solution, mass_ratio, model_params)[1]
Newt_waveform = compute_waveform(dt_data, Newtonian_solution, mass_ratio, model_params)[1]

df_predicted_trajectories = DataFrame(time = extended_tsteps,
                         true_orbit_x = true_orbit[1,:],
                         true_orbit_y = true_orbit[2,:],
                         pred_orbit_x = pred_orbit[1,:],
                         pred_orbit_y = pred_orbit[2,:],
                         Newt_orbit_x = Newt_orbit[1,:],
                         Newt_orbit_y = Newt_orbit[2,:])

df_predicted_waveforms = DataFrame(time = extended_tsteps,
                                true_waveform = true_waveform,
                                pred_waveform = pred_waveform,
                                Newt_waveform = Newt_waveform)

## Save data
save_data = false

if save_data == true
    CSV.write("./output/EMRI_learned_trajectories.csv", df_learned_trajectories)
    CSV.write("./output/EMRI_predicted_trajectories.csv", df_predicted_trajectories)
    CSV.write("./output/EMRI_learned_waveforms.csv", df_learned_waveforms)
    CSV.write("./output/EMRI_predicted_waveforms.csv", df_predicted_waveforms)
end
